@@ -8,7 +8,7 @@
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/backend/utils/adt/varlena.c,v 1.92.2.1 2003/04/23 18:19:23 tgl Exp $
+ *	  $Header: /cvsroot/pgsql/src/backend/utils/adt/varlena.c,v 1.92.2.3 2004/02/01 04:05:13 joe Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -174,10 +174,10 @@ byteaout(PG_FUNCTION_ARGS)
 	{
 		if (*vp == '\\')
 			len += 2;
-		else if (isprint((unsigned char) *vp))
-			len++;
-		else
+		else if ((unsigned char) *vp < 0x20 || (unsigned char) *vp > 0x7e)
 			len += 4;
+		else
+			len++;
 	}
 	rp = result = (char *) palloc(len);
 	vp = vlena->vl_dat;
@@ -188,9 +188,7 @@ byteaout(PG_FUNCTION_ARGS)
 			*rp++ = '\\';
 			*rp++ = '\\';
 		}
-		else if (isprint((unsigned char) *vp))
-			*rp++ = *vp;
-		else
+		else if ((unsigned char) *vp < 0x20 || (unsigned char) *vp > 0x7e)
 		{
 			val = *vp;
 			rp[0] = '\\';
@@ -201,6 +199,8 @@ byteaout(PG_FUNCTION_ARGS)
 			rp[1] = DIG(val & 03);
 			rp += 4;
 		}
+		else
+			*rp++ = *vp;
 	}
 	*rp = '\0';
 	PG_RETURN_CSTRING(result);
@@ -665,9 +665,6 @@ text_position(Datum str, Datum search_str, int matchnum)
 	len1 = (VARSIZE(t1) - VARHDRSZ);
 	len2 = (VARSIZE(t2) - VARHDRSZ);
 
-	/* no use in searching str past point where search_str will fit */
-	px = (len1 - len2);
-
 	if (eml == 1)				/* simple case - single byte encoding */
 	{
 		char	   *p1,
@@ -676,6 +673,9 @@ text_position(Datum str, Datum search_str, int matchnum)
 		p1 = VARDATA(t1);
 		p2 = VARDATA(t2);
 
+		/* no use in searching str past point where search_str will fit */
+		px = (len1 - len2);
+
 		for (p = 0; p <= px; p++)
 		{
 			if ((*p2 == *p1) && (strncmp(p1, p2, len2) == 0))
@@ -703,6 +703,9 @@ text_position(Datum str, Datum search_str, int matchnum)
 		(void) pg_mb2wchar_with_len((unsigned char *) VARDATA(t2), p2, len2);
 		len2 = pg_wchar_strlen(p2);
 
+		/* no use in searching str past point where search_str will fit */
+		px = (len1 - len2);
+
 		for (p = 0; p <= px; p++)
 		{
 			if ((*p2 == *p1) && (pg_wchar_strncmp(p1, p2, len2) == 0))
