@@ -4,10 +4,10 @@
  *	  include file for the bootstrapping code
  *
  *
- * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
- * $Id: bootstrap.h,v 1.26 2001/11/05 17:46:31 momjian Exp $
+ * $Id: bootstrap.h,v 1.30 2002/06/20 20:29:43 momjian Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -31,14 +31,12 @@ typedef struct hashnode
 } hashnode;
 
 
-extern Relation reldesc;
+extern Relation boot_reldesc;
 extern Form_pg_attribute attrtypes[MAXATTR];
 extern int	numattr;
-extern int	DebugMode;
-
 extern int	BootstrapMain(int ac, char *av[]);
 
-extern void index_register(char *heap, char *ind, IndexInfo *indexInfo);
+extern void index_register(Oid heap, Oid ind, IndexInfo *indexInfo);
 
 extern void err_out(void);
 extern void InsertOneTuple(Oid objectid);
